@@ -15,6 +15,7 @@
 #include "ntp_stdlib.h"
 
 #include <stdio.h>
+#include <stddef.h>
 #include <signal.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
@@ -39,6 +40,8 @@
  */
 static int client_v6_capable = 0;   /* the client can handle longer messages */
 
+#define v6sizeof(type)	(client_v6_capable ? sizeof(type) : v4sizeof(type))
+
 struct req_proc {
 	short request_code;	/* defined request code */
 	short needs_auth;	/* true when authentication needed */
@@ -95,7 +98,9 @@ static	void	req_clr_trap	P((struct sockaddr_storage *, struct interface *, struc
 static	void	do_setclr_trap	P((struct sockaddr_storage *, struct interface *, struct req_pkt *, int));
 static	void	set_request_keyid P((struct sockaddr_storage *, struct interface *, struct req_pkt *));
 static	void	set_control_keyid P((struct sockaddr_storage *, struct interface *, struct req_pkt *));
-static	void	get_ctl_stats P((struct sockaddr_storage *, struct interface *, struct req_pkt *));
+static	void	get_ctl_stats   P((struct sockaddr_storage *, struct interface *, struct req_pkt *));
+static	void	get_if_stats    P((struct sockaddr_storage *, struct interface *, struct req_pkt *));
+static	void	do_if_reload    P((struct sockaddr_storage *, struct interface *, struct req_pkt *));
 #ifdef KERNEL_PLL
 static	void	get_kernel_info P((struct sockaddr_storage *, struct interface *, struct req_pkt *));
 #endif /* KERNEL_PLL */
@@ -113,9 +118,9 @@ static	void	get_clkbug_info P((struct sockaddr_storage *, struct interface *, st
 static	struct req_proc ntp_codes[] = {
 	{ REQ_PEER_LIST,	NOAUTH,	0, 0,	peer_list },
 	{ REQ_PEER_LIST_SUM,	NOAUTH,	0, 0,	peer_list_sum },
-	{ REQ_PEER_INFO,    NOAUTH, (sizeof(struct info_peer_list) - ALT_LONG),
+	{ REQ_PEER_INFO,    NOAUTH, v4sizeof(struct info_peer_list),
 				sizeof(struct info_peer_list), peer_info},
-	{ REQ_PEER_STATS,   NOAUTH, (sizeof(struct info_peer_list) - ALT_LONG),
+	{ REQ_PEER_STATS,   NOAUTH, v4sizeof(struct info_peer_list),
 				sizeof(struct info_peer_list), peer_stats},
 	{ REQ_SYS_INFO,		NOAUTH,	0, 0,	sys_info },
 	{ REQ_SYS_STATS,	NOAUTH,	0, 0,	sys_stats },
@@ -123,34 +128,34 @@ static	struct req_proc ntp_codes[] = {
 	{ REQ_MEM_STATS,	NOAUTH,	0, 0,	mem_stats },
 	{ REQ_LOOP_INFO,	NOAUTH,	0, 0,	loop_info },
 	{ REQ_TIMER_STATS,	NOAUTH,	0, 0,	timer_stats },
-	{ REQ_CONFIG,	    AUTH, (sizeof(struct conf_peer) - ALT_LONG),
+	{ REQ_CONFIG,	    AUTH, v4sizeof(struct conf_peer),
 				sizeof(struct conf_peer), do_conf },
-	{ REQ_UNCONFIG,	    AUTH, (sizeof(struct conf_unpeer) - ALT_INT),
+	{ REQ_UNCONFIG,	    AUTH, v4sizeof(struct conf_unpeer),
 				sizeof(struct conf_unpeer), do_unconf },
 	{ REQ_SET_SYS_FLAG, AUTH, sizeof(struct conf_sys_flags),
 				sizeof(struct conf_sys_flags), set_sys_flag },
 	{ REQ_CLR_SYS_FLAG, AUTH, sizeof(struct conf_sys_flags), 
 				sizeof(struct conf_sys_flags),  clr_sys_flag },
 	{ REQ_GET_RESTRICT,	NOAUTH,	0, 0,	list_restrict },
-	{ REQ_RESADDFLAGS, AUTH, (sizeof(struct conf_restrict) - ALT_LONG),
+	{ REQ_RESADDFLAGS, AUTH, v4sizeof(struct conf_restrict),
 				sizeof(struct conf_restrict), do_resaddflags },
-	{ REQ_RESSUBFLAGS, AUTH, (sizeof(struct conf_restrict) - ALT_LONG),
+	{ REQ_RESSUBFLAGS, AUTH, v4sizeof(struct conf_restrict),
 				sizeof(struct conf_restrict), do_ressubflags },
-	{ REQ_UNRESTRICT, AUTH, (sizeof(struct conf_restrict) - ALT_LONG),
+	{ REQ_UNRESTRICT, AUTH, v4sizeof(struct conf_restrict),
 				sizeof(struct conf_restrict), do_unrestrict },
 	{ REQ_MON_GETLIST,	NOAUTH,	0, 0,	mon_getlist_0 },
 	{ REQ_MON_GETLIST_1,	NOAUTH,	0, 0,	mon_getlist_1 },
 	{ REQ_RESET_STATS, AUTH, sizeof(struct reset_flags), 0, reset_stats },
-	{ REQ_RESET_PEER,  AUTH, (sizeof(struct conf_unpeer) - ALT_INT),
+	{ REQ_RESET_PEER,  AUTH, v4sizeof(struct conf_unpeer),
 				sizeof(struct conf_unpeer), reset_peer },
 	{ REQ_REREAD_KEYS,	AUTH,	0, 0,	do_key_reread },
 	{ REQ_TRUSTKEY,   AUTH, sizeof(u_long), sizeof(u_long), trust_key },
 	{ REQ_UNTRUSTKEY, AUTH, sizeof(u_long), sizeof(u_long), untrust_key },
 	{ REQ_AUTHINFO,		NOAUTH,	0, 0,	get_auth_info },
 	{ REQ_TRAPS,		NOAUTH, 0, 0,	req_get_traps },
-	{ REQ_ADD_TRAP,	AUTH, (sizeof(struct conf_trap) - ALT_LONG),
+	{ REQ_ADD_TRAP,	AUTH, v4sizeof(struct conf_trap),
 				sizeof(struct conf_trap), req_set_trap },
-	{ REQ_CLR_TRAP,	AUTH, (sizeof(struct conf_trap) - ALT_LONG),
+	{ REQ_CLR_TRAP,	AUTH, v4sizeof(struct conf_trap),
 				sizeof(struct conf_trap), req_clr_trap },
 	{ REQ_REQUEST_KEY, AUTH, sizeof(u_long), sizeof(u_long), 
 				set_request_keyid },
@@ -168,6 +173,9 @@ static	struct req_proc ntp_codes[] = {
 	{ REQ_GET_CLKBUGINFO, NOAUTH, sizeof(u_int32), sizeof(u_int32),
 				get_clkbug_info },
 #endif
+	{ REQ_IF_STATS,		AUTH, 0, 0,	get_if_stats },
+	{ REQ_IF_RELOAD,        AUTH, 0, 0,	do_if_reload },
+
 	{ NO_REQUEST,		NOAUTH,	0, 0,	0 }
 };
 
@@ -401,6 +409,7 @@ process_private(
 	int mod_okay
 	)
 {
+	static u_long quiet_until;
 	struct req_pkt *inpkt;
 	struct req_pkt_tail *tailinpkt;
 	struct sockaddr_storage *srcadr;
@@ -436,8 +445,14 @@ process_private(
 	    || (++ec, INFO_MBZ(inpkt->mbz_itemsize) != 0)
 	    || (++ec, rbufp->recv_length < REQ_LEN_HDR)
 		) {
-		msyslog(LOG_ERR, "process_private: INFO_ERR_FMT: test %d failed, pkt from %s", ec, stoa(srcadr));
-		req_ack(srcadr, inter, inpkt, INFO_ERR_FMT);
+		NLOG(NLOG_SYSEVENT)
+			if (current_time >= quiet_until) {
+				msyslog(LOG_ERR,
+					"process_private: drop test %d"
+					" failed, pkt from %s",
+					ec, stoa(srcadr));
+				quiet_until = current_time + 60;
+			}
 		return;
 	}
 
@@ -488,21 +503,25 @@ process_private(
 	 */
 	temp_size = INFO_ITEMSIZE(inpkt->mbz_itemsize);
 	if ((temp_size != proc->sizeofitem &&
-	    temp_size != proc->v6_sizeofitem) &&
+	     temp_size != proc->v6_sizeofitem) &&
 	    !(inpkt->implementation == IMPL_XNTPD &&
-	    inpkt->request == REQ_CONFIG &&
-	    temp_size == sizeof(struct old_conf_peer))) {
+	      inpkt->request == REQ_CONFIG &&
+	      temp_size == sizeof(struct old_conf_peer))) {
+#ifdef DEBUG
 		if (debug > 2)
 			printf("process_private: wrong item size, received %d, should be %d or %d\n",
 			    temp_size, proc->sizeofitem, proc->v6_sizeofitem);
+#endif
 		req_ack(srcadr, inter, inpkt, INFO_ERR_FMT);
 		return;
 	}
 	if ((proc->sizeofitem != 0) &&
 	    ((temp_size * INFO_NITEMS(inpkt->err_nitems)) >
 	    (rbufp->recv_length - REQ_LEN_HDR))) {
+#ifdef DEBUG
 		if (debug > 2)
 			printf("process_private: not enough data\n");
+#endif
 		req_ack(srcadr, inter, inpkt, INFO_ERR_FMT);
 		return;
 	}
@@ -531,10 +550,10 @@ process_private(
 		l_fp ftmp;
 		double dtemp;
 	
-		if (rbufp->recv_length < (REQ_LEN_HDR +
+		if (rbufp->recv_length < (int)((REQ_LEN_HDR +
 		    (INFO_ITEMSIZE(inpkt->mbz_itemsize) *
 		    INFO_NITEMS(inpkt->err_nitems))
-		    + sizeof(struct req_pkt_tail))) {
+		    + sizeof(struct req_pkt_tail)))) {
 			req_ack(srcadr, inter, inpkt, INFO_ERR_FMT);
 		}
 		tailinpkt = (struct req_pkt_tail *)((char *)&rbufp->recv_pkt +
@@ -549,9 +568,14 @@ process_private(
 #ifdef DEBUG
 			if (debug > 4)
 			    printf("failed auth %d info_auth_keyid %lu pkt keyid %lu\n",
-				    INFO_IS_AUTH(inpkt->auth_seq),
-				    (u_long)info_auth_keyid,
-				    (u_long)ntohl(tailinpkt->keyid));
+				   INFO_IS_AUTH(inpkt->auth_seq),
+				   (u_long)info_auth_keyid,
+				   (u_long)ntohl(tailinpkt->keyid));
+			msyslog(LOG_DEBUG,
+				"process_private: failed auth %d info_auth_keyid %lu pkt keyid %lu\n",
+				INFO_IS_AUTH(inpkt->auth_seq),
+				(u_long)info_auth_keyid,
+				(u_long)ntohl(tailinpkt->keyid));
 #endif
 			req_ack(srcadr, inter, inpkt, INFO_ERR_AUTH);
 			return;
@@ -571,6 +595,9 @@ process_private(
 #ifdef DEBUG
 			if (debug > 4)
 			    printf("failed auth mod_okay %d\n", mod_okay);
+			msyslog(LOG_DEBUG,
+				"process_private: failed auth mod_okay %d\n",
+				mod_okay);
 #endif
 			req_ack(srcadr, inter, inpkt, INFO_ERR_AUTH);
 			return;
@@ -587,6 +614,10 @@ process_private(
 			/*
 			 * He's a loser.  Tell him.
 			 */
+#ifdef DEBUG
+			if (debug > 4)
+			    printf("xmit/rcv timestamp delta > INFO_TS_MAXSKEW\n");
+#endif
 			req_ack(srcadr, inter, inpkt, INFO_ERR_AUTH);
 			return;
 		}
@@ -597,6 +628,10 @@ process_private(
 		if (!authdecrypt(info_auth_keyid, (u_int32 *)inpkt,
 		    rbufp->recv_length - sizeof(struct req_pkt_tail) +
 		    REQ_LEN_HDR, sizeof(struct req_pkt_tail) - REQ_LEN_HDR)) {
+#ifdef DEBUG
+			if (debug > 4)
+			    printf("authdecrypt failed\n");
+#endif
 			req_ack(srcadr, inter, inpkt, INFO_ERR_AUTH);
 			return;
 		}
@@ -628,15 +663,10 @@ peer_list(
 	register struct peer *pp;
 	register int i;
 	register int skip = 0;
-	int offset;
 
-	if (client_v6_capable)
-		offset = 0;
-	else
-		offset = ALT_LONG;
 	ip = (struct info_peer_list *)prepare_pkt(srcadr, inter, inpkt,
-	    sizeof(struct info_peer_list) - offset);
-	for (i = 0; i < HASH_SIZE && ip != 0; i++) {
+	    v6sizeof(struct info_peer_list));
+	for (i = 0; i < NTP_HASH_SIZE && ip != 0; i++) {
 		pp = peer_hash[i];
 		while (pp != 0 && ip != 0) {
 			if (pp->srcadr.ss_family == AF_INET6) {
@@ -691,19 +721,14 @@ peer_list_sum(
 	register int i;
 	l_fp ltmp;
 	register int skip;
-	int offset;
 
 #ifdef DEBUG
 	if (debug > 2)
 	    printf("wants peer list summary\n");
 #endif
-	if (client_v6_capable)
-		offset = 0;
-	else
-		offset = ALT_2ADDR_LONG;
 	ips = (struct info_peer_summary *)prepare_pkt(srcadr, inter, inpkt,
-	    sizeof(struct info_peer_summary) - offset);
-	for (i = 0; i < HASH_SIZE && ips != 0; i++) {
+	    v6sizeof(struct info_peer_summary));
+	for (i = 0; i < NTP_HASH_SIZE && ips != 0; i++) {
 		pp = peer_hash[i];
 		while (pp != 0 && ips != 0) {
 #ifdef DEBUG
@@ -718,7 +743,10 @@ peer_list_sum(
 				if (client_v6_capable) {
 					ips->srcadr6 = GET_INADDR6(pp->srcadr);
 					ips->v6_flag = 1;
-					ips->dstadr6 = GET_INADDR6(pp->dstadr->sin);
+					if (pp->dstadr)
+						ips->dstadr6 = GET_INADDR6(pp->dstadr->sin);
+					else
+						memset(&ips->dstadr6, 0, sizeof(ips->dstadr6));
 					skip = 0;
 				} else {
 					skip = 1;
@@ -728,16 +756,23 @@ peer_list_sum(
 				ips->srcadr = GET_INADDR(pp->srcadr);
 				if (client_v6_capable)
 					ips->v6_flag = 0;
-				ips->dstadr = (pp->processed) ?
-					pp->cast_flags == MDF_BCAST ?
-					GET_INADDR(pp->dstadr->bcast):
-					pp->cast_flags ?
-					GET_INADDR(pp->dstadr->sin) ?
-					GET_INADDR(pp->dstadr->sin):
-					GET_INADDR(pp->dstadr->bcast):
-					1 : 5;
+/* XXX PDM This code is buggy. Need to replace with a straightforward assignment */
+				
+				if (pp->dstadr)
+					ips->dstadr = (pp->processed) ?
+						pp->cast_flags == MDF_BCAST ?
+						GET_INADDR(pp->dstadr->bcast):
+						pp->cast_flags ?
+						GET_INADDR(pp->dstadr->sin) ?
+						GET_INADDR(pp->dstadr->sin):
+						GET_INADDR(pp->dstadr->bcast):
+						1 : GET_INADDR(pp->dstadr->sin);
+				else
+						memset(&ips->dstadr, 0, sizeof(ips->dstadr));
+
 				skip = 0;
 			}
+			
 			if (!skip){ 
 				ips->srcport = NSRCPORT(&pp->srcadr);
 				ips->stratum = pp->stratum;
@@ -765,7 +800,7 @@ peer_list_sum(
 				ips->delay = HTONS_FP(DTOFP(pp->delay));
 				DTOLFP(pp->offset, &ltmp);
 				HTONL_FP(&ltmp, &ips->offset);
-				ips->dispersion = HTONS_FP(DTOUFP(pp->disp));
+				ips->dispersion = HTONS_FP(DTOUFP(SQRT(pp->disp)));
 			}	
 			pp = pp->next; 
 			ips = (struct info_peer_summary *)more_pkt();
@@ -790,7 +825,6 @@ peer_info (
 	register struct info_peer *ip;
 	register int items;
 	register int i, j;
-	int offset;
 	struct sockaddr_storage addr;
 	extern struct peer *sys_peer;
 	l_fp ltmp;
@@ -799,12 +833,8 @@ peer_info (
 	items = INFO_NITEMS(inpkt->err_nitems);
 	ipl = (struct info_peer_list *) inpkt->data;
 
-	if (client_v6_capable)
-		offset = 0;
-	else
-		offset = ALT_2ADDR_LONG;
 	ip = (struct info_peer *)prepare_pkt(srcadr, inter, inpkt,
-	    sizeof(struct info_peer) - offset);
+	    v6sizeof(struct info_peer));
 	while (items-- > 0 && ip != 0) {
 		memset((char *)&addr, 0, sizeof(addr));
 		NSRCPORT(&addr) = ipl->port;
@@ -822,20 +852,29 @@ peer_info (
 		if ((pp = findexistingpeer(&addr, (struct peer *)0, -1)) == 0)
 		    continue;
 		if (pp->srcadr.ss_family == AF_INET6) {
-			ip->dstadr6 = pp->cast_flags == MDF_BCAST ?
-				GET_INADDR6(pp->dstadr->bcast) :
-				GET_INADDR6(pp->dstadr->sin);
+			if (pp->dstadr)
+				ip->dstadr6 = pp->cast_flags == MDF_BCAST ?
+					GET_INADDR6(pp->dstadr->bcast) :
+					GET_INADDR6(pp->dstadr->sin);
+			else
+				memset(&ip->dstadr6, 0, sizeof(ip->dstadr6));
+
 			ip->srcadr6 = GET_INADDR6(pp->srcadr);
 			ip->v6_flag = 1;
 		} else {
-			ip->dstadr = (pp->processed) ?
-				pp->cast_flags == MDF_BCAST ?
-				GET_INADDR(pp->dstadr->bcast):
-				pp->cast_flags ?
-				GET_INADDR(pp->dstadr->sin) ?
-				GET_INADDR(pp->dstadr->sin):
-				GET_INADDR(pp->dstadr->bcast):
-				2 : 6;
+/* XXX PDM This code is buggy. Need to replace with a straightforward assignment */
+			if (pp->dstadr)
+				ip->dstadr = (pp->processed) ?
+					pp->cast_flags == MDF_BCAST ?
+					GET_INADDR(pp->dstadr->bcast):
+					pp->cast_flags ?
+					GET_INADDR(pp->dstadr->sin) ?
+					GET_INADDR(pp->dstadr->sin):
+					GET_INADDR(pp->dstadr->bcast):
+					2 : GET_INADDR(pp->dstadr->sin);
+			else
+				memset(&ip->dstadr, 0, sizeof(ip->dstadr));
+
 			ip->srcadr = GET_INADDR(pp->srcadr);
 			if (client_v6_capable)
 				ip->v6_flag = 0;
@@ -887,8 +926,8 @@ peer_info (
 			ip->filtdelay[i] = HTONS_FP(DTOFP(pp->filter_delay[j]));
 			DTOLFP(pp->filter_offset[j], &ltmp);
 			HTONL_FP(&ltmp, &ip->filtoffset[i]);
-			ip->order[i] = (pp->filter_nextpt+NTP_SHIFT-1)
-				- pp->filter_order[i];
+			ip->order[i] = (u_char)((pp->filter_nextpt+NTP_SHIFT-1)
+				- pp->filter_order[i]);
 			if (ip->order[i] >= NTP_SHIFT)
 			    ip->order[i] -= NTP_SHIFT;
 		}
@@ -917,19 +956,17 @@ peer_stats (
 	register struct peer *pp;
 	register struct info_peer_stats *ip;
 	register int items;
-	int offset;
 	struct sockaddr_storage addr;
 	extern struct peer *sys_peer;
 
-	printf("peer_stats: called\n");
+#ifdef DEBUG
+	if (debug)
+	     printf("peer_stats: called\n");
+#endif
 	items = INFO_NITEMS(inpkt->err_nitems);
 	ipl = (struct info_peer_list *) inpkt->data;
-	if (client_v6_capable)
-		offset = 0;
-	else
-		offset = ALT_2ADDR_LONG;
 	ip = (struct info_peer_stats *)prepare_pkt(srcadr, inter, inpkt,
-	    sizeof(struct info_peer_stats) - offset);
+	    v6sizeof(struct info_peer_stats));
 	while (items-- > 0 && ip != 0) {
 		memset((char *)&addr, 0, sizeof(addr));
 		NSRCPORT(&addr) = ipl->port;
@@ -943,30 +980,44 @@ peer_stats (
 #ifdef HAVE_SA_LEN_IN_STRUCT_SOCKADDR
 		addr.ss_len = SOCKLEN(&addr);
 #endif
-		printf("peer_stats: looking for %s, %d, %d\n", stoa(&addr),
+#ifdef DEBUG
+		if (debug)
+		    printf("peer_stats: looking for %s, %d, %d\n", stoa(&addr),
 		    ipl->port, ((struct sockaddr_in6 *)&addr)->sin6_port);
+#endif
 		ipl = (struct info_peer_list *)((char *)ipl +
 		    INFO_ITEMSIZE(inpkt->mbz_itemsize));
 
 		if ((pp = findexistingpeer(&addr, (struct peer *)0, -1)) == 0)
 		    continue;
-		printf("peer_stats: found %s\n", stoa(&addr));
+#ifdef DEBUG
+		if (debug)
+		     printf("peer_stats: found %s\n", stoa(&addr));
+#endif
 		if (pp->srcadr.ss_family == AF_INET) {
-			ip->dstadr = (pp->processed) ?
-				pp->cast_flags == MDF_BCAST ?
-				GET_INADDR(pp->dstadr->bcast):
-				pp->cast_flags ?
-				GET_INADDR(pp->dstadr->sin) ?
-				GET_INADDR(pp->dstadr->sin):
-				GET_INADDR(pp->dstadr->bcast):
-				3 : 7;
+			if (pp->dstadr)
+				ip->dstadr = (pp->processed) ?
+					pp->cast_flags == MDF_BCAST ?
+					GET_INADDR(pp->dstadr->bcast):
+					pp->cast_flags ?
+					GET_INADDR(pp->dstadr->sin) ?
+					GET_INADDR(pp->dstadr->sin):
+					GET_INADDR(pp->dstadr->bcast):
+					3 : 7;
+			else
+				memset(&ip->dstadr, 0, sizeof(ip->dstadr));
+			
 			ip->srcadr = GET_INADDR(pp->srcadr);
 			if (client_v6_capable)
 				ip->v6_flag = 0;
 		} else {
-			ip->dstadr6 = pp->cast_flags == MDF_BCAST ?
-				GET_INADDR6(pp->dstadr->bcast):
-				GET_INADDR6(pp->dstadr->sin);
+			if (pp->dstadr)
+				ip->dstadr6 = pp->cast_flags == MDF_BCAST ?
+					GET_INADDR6(pp->dstadr->bcast):
+					GET_INADDR6(pp->dstadr->sin);
+			else
+				memset(&ip->dstadr6, 0, sizeof(ip->dstadr6));
+			
 			ip->srcadr6 = GET_INADDR6(pp->srcadr);
 			ip->v6_flag = 1;
 		}	
@@ -984,10 +1035,13 @@ peer_stats (
 		    ip->flags |= INFO_FLAG_PREFER;
 		if (pp->flags & FLAG_BURST)
 		    ip->flags |= INFO_FLAG_BURST;
+		if (pp->flags & FLAG_IBURST)
+		    ip->flags |= INFO_FLAG_IBURST;
 		if (pp->status == CTL_PST_SEL_SYNCCAND)
 		    ip->flags |= INFO_FLAG_SEL_CANDIDATE;
 		if (pp->status >= CTL_PST_SEL_SYSPEER)
 		    ip->flags |= INFO_FLAG_SHORTLIST;
+		ip->flags = htons(ip->flags);
 		ip->timereceived = htonl((u_int32)(current_time - pp->timereceived));
 		ip->timetosend = htonl(pp->nextdate - current_time);
 		ip->timereachable = htonl((u_int32)(current_time - pp->timereachable));
@@ -1016,35 +1070,10 @@ sys_info(
 	)
 {
 	register struct info_sys *is;
-	int offset;
 
-	/*
-	 * Importations from the protocol module
-	 */
-	extern u_char sys_leap;
-	extern u_char sys_stratum;
-	extern s_char sys_precision;
-	extern double sys_rootdelay;
-	extern double sys_rootdispersion;
-	extern u_int32 sys_refid;
-	extern l_fp sys_reftime;
-	extern u_char sys_poll;
-	extern struct peer *sys_peer;
-	extern int sys_bclient;
-	extern double sys_bdelay;
-	extern l_fp sys_authdelay;
-	extern double clock_stability;
-	extern double sys_jitter;
-
-	if (client_v6_capable)
-		offset = 0;
-	else 
-		offset = ALT_LONG;
 	is = (struct info_sys *)prepare_pkt(srcadr, inter, inpkt,
-	    sizeof(struct info_sys) - offset);
+	    v6sizeof(struct info_sys));
 
-	is->peer = 0;
-	is->peer_mode = 0;
 	if (sys_peer != 0) {
 		if (sys_peer->srcadr.ss_family == AF_INET) {
 			is->peer = GET_INADDR(sys_peer->srcadr);
@@ -1055,6 +1084,12 @@ sys_info(
 			is->v6_flag = 1;
 		}
 		is->peer_mode = sys_peer->hmode;
+	} else {
+		is->peer = 0;
+		if (client_v6_capable) {
+			is->v6_flag = 0;
+		}
+		is->peer_mode = 0;
 	}
 
 	is->leap = sys_leap;
@@ -1063,7 +1098,7 @@ sys_info(
 	is->rootdelay = htonl(DTOFP(sys_rootdelay));
 	is->rootdispersion = htonl(DTOUFP(sys_rootdispersion));
 	is->frequency = htonl(DTOFP(sys_jitter));
-	is->stability = htonl(DTOUFP(clock_stability * 1e6));
+	is->stability = htonl(DTOUFP(clock_stability));
 	is->refid = sys_refid;
 	HTONL_FP(&sys_reftime, &is->reftime);
 
@@ -1145,7 +1180,7 @@ mem_stats(
 	/*
 	 * Importations from the peer module
 	 */
-	extern int peer_hash_count[HASH_SIZE];
+	extern int peer_hash_count[NTP_HASH_SIZE];
 	extern int peer_free_count;
 	extern u_long peer_timereset;
 	extern u_long findpeer_calls;
@@ -1163,7 +1198,7 @@ mem_stats(
 	ms->allocations = htonl((u_int32)peer_allocations);
 	ms->demobilizations = htonl((u_int32)peer_demobilizations);
 
-	for (i = 0; i < HASH_SIZE; i++) {
+	for (i = 0; i < NTP_HASH_SIZE; i++) {
 		if (peer_hash_count[i] > 255)
 		    ms->hashcount[i] = 255;
 		else
@@ -1264,7 +1299,7 @@ loop_info(
 	extern double last_offset;
 	extern double drift_comp;
 	extern int tc_counter;
-	extern u_long last_time;
+	extern u_long sys_clocktime;
 
 	li = (struct info_loop *)prepare_pkt(srcadr, inter, inpkt,
 	    sizeof(struct info_loop));
@@ -1274,7 +1309,7 @@ loop_info(
 	DTOLFP(drift_comp * 1e6, &ltmp);
 	HTONL_FP(&ltmp, &li->drift_comp);
 	li->compliance = htonl((u_int32)(tc_counter));
-	li->watchdog_timer = htonl((u_int32)(current_time - last_time));
+	li->watchdog_timer = htonl((u_int32)(current_time - sys_clocktime));
 
 	(void) more_pkt();
 	flush_pkt();
@@ -1291,6 +1326,7 @@ do_conf(
 	struct req_pkt *inpkt
 	)
 {
+	static u_long soonest_ifrescan_time = 0;
 	int items;
 	u_int fl;
 	struct conf_peer *cp; 
@@ -1317,7 +1353,7 @@ do_conf(
 		    && temp_cp.hmode != MODE_BROADCAST)
 		    fl = 1;
 		if (temp_cp.flags & ~(CONF_FLAG_AUTHENABLE | CONF_FLAG_PREFER
-				  | CONF_FLAG_BURST | CONF_FLAG_SKEY))
+				  | CONF_FLAG_BURST | CONF_FLAG_IBURST | CONF_FLAG_SKEY))
 		    fl = 1;
 		cp = (struct conf_peer *)
 		    ((char *)cp + INFO_ITEMSIZE(inpkt->mbz_itemsize));
@@ -1341,13 +1377,16 @@ do_conf(
 
 		fl = 0;
 		if (temp_cp.flags & CONF_FLAG_AUTHENABLE)
-		    fl |= FLAG_AUTHENABLE;
+			fl |= FLAG_AUTHENABLE;
 		if (temp_cp.flags & CONF_FLAG_PREFER)
-		    fl |= FLAG_PREFER;
+			fl |= FLAG_PREFER;
 		if (temp_cp.flags & CONF_FLAG_BURST)
 		    fl |= FLAG_BURST;
+		if (temp_cp.flags & CONF_FLAG_IBURST)
+		    fl |= FLAG_IBURST;
 		if (temp_cp.flags & CONF_FLAG_SKEY)
 			fl |= FLAG_SKEY;
+		
 		if (client_v6_capable && temp_cp.v6_flag != 0) {
 			peeraddr.ss_family = AF_INET6;
 			GET_INADDR6(peeraddr) = temp_cp.peeraddr6; 
@@ -1381,6 +1420,23 @@ do_conf(
 			req_ack(srcadr, inter, inpkt, INFO_ERR_NODATA);
 			return;
 		}
+
+		/*
+		 * ntp_intres.c uses REQ_CONFIG/doconf() to add each
+		 * server after its name is resolved.  If we have been
+		 * disconnected from the network, it may notice the
+		 * network has returned and add the first server while
+		 * the relevant interface is still disabled, awaiting
+		 * the next interface rescan.  To get things moving
+		 * more quickly, trigger an interface scan now, except
+		 * if we have done so in the last half minute.
+		 */
+		if (soonest_ifrescan_time < current_time) {
+			soonest_ifrescan_time = current_time + 30;
+			timer_interfacetimeout(current_time);
+			DPRINTF(1, ("do_conf triggering interface rescan\n"));
+		}
+
 		cp = (struct conf_peer *)
 		    ((char *)cp + INFO_ITEMSIZE(inpkt->mbz_itemsize));
 	}
@@ -1529,7 +1585,10 @@ do_unconf(
 #endif
 		found = 0;
 		peer = (struct peer *)0;
-		printf("searching for %s\n", stoa(&peeraddr));
+#ifdef DEBUG
+		if (debug)
+		     printf("searching for %s\n", stoa(&peeraddr));
+#endif
 		while (!found) {
 			peer = findexistingpeer(&peeraddr, peer, -1);
 			if (peer == (struct peer *)0)
@@ -1618,7 +1677,9 @@ setclr_flags(
 	)
 {
 	register u_int flags;
+	int prev_kern_enable;
 
+	prev_kern_enable = kern_enable;
 	if (INFO_NITEMS(inpkt->err_nitems) > 1) {
 		msyslog(LOG_ERR, "setclr_flags: err_nitems > 1");
 		req_ack(srcadr, inter, inpkt, INFO_ERR_FMT);
@@ -1626,7 +1687,8 @@ setclr_flags(
 	}
 
 	flags = ((struct conf_sys_flags *)inpkt->data)->flags;
-
+	flags = ntohl(flags);
+	
 	if (flags & ~(SYS_FLAG_BCLIENT | SYS_FLAG_PPS |
 		      SYS_FLAG_NTP | SYS_FLAG_KERNEL | SYS_FLAG_MONITOR |
 		      SYS_FLAG_FILEGEN | SYS_FLAG_AUTH | SYS_FLAG_CAL)) {
@@ -1656,6 +1718,12 @@ setclr_flags(
 	if (flags & SYS_FLAG_CAL)
 		proto_config(PROTO_CAL, set, 0., NULL);
 	req_ack(srcadr, inter, inpkt, INFO_OKAY);
+
+	/* Reset the kernel ntp parameters if the kernel flag changed. */
+	if (prev_kern_enable && !kern_enable)
+	     	loop_config(LOOP_KERN_CLEAR, 0.0);
+	if (!prev_kern_enable && kern_enable)
+	     	loop_config(LOOP_DRIFTCOMP, drift_comp);
 }
 
 
@@ -1672,19 +1740,14 @@ list_restrict(
 	register struct info_restrict *ir;
 	register struct restrictlist *rl;
 	register struct restrictlist6 *rl6;
-	int offset;
 
 #ifdef DEBUG
 	if (debug > 2)
 	    printf("wants restrict list summary\n");
 #endif
-	if (client_v6_capable) 
-		offset = 0;
-	else
-		offset = ALT_2ADDR_LONG;	
 
 	ir = (struct info_restrict *)prepare_pkt(srcadr, inter, inpkt,
-	    sizeof(struct info_restrict) - offset);
+	    v6sizeof(struct info_restrict));
 	
 	for (rl = restrictlist; rl != 0 && ir != 0; rl = rl->next) {
 		ir->addr = htonl(rl->addr);
@@ -1783,6 +1846,8 @@ do_restrict(
 	cr = (struct conf_restrict *)inpkt->data;
 
 	bad = 0;
+	cr->flags = ntohs(cr->flags);
+	cr->mflags = ntohs(cr->mflags);
 	while (items-- > 0 && !bad) {
 		if (cr->mflags & ~(RESM_NTPONLY))
 		    bad |= 1;
@@ -1847,7 +1912,6 @@ mon_getlist_0(
 {
 	register struct info_monitor *im;
 	register struct mon_data *md;
-	int offset;
 	extern struct mon_data mon_mru_list;
 	extern int mon_enabled;
 
@@ -1859,16 +1923,12 @@ mon_getlist_0(
 		req_ack(srcadr, inter, inpkt, INFO_ERR_NODATA);
 		return;
 	}
-	if (client_v6_capable)
-		offset = 0;
-	else
-		offset = ALT_LONG;
 	im = (struct info_monitor *)prepare_pkt(srcadr, inter, inpkt,
-	    sizeof(struct info_monitor) - offset);
+	    v6sizeof(struct info_monitor));
 	for (md = mon_mru_list.mru_next; md != &mon_mru_list && im != 0;
 	     md = md->mru_next) {
 		im->lasttime = htonl((u_int32)md->avg_interval);
-		im->firsttime = htonl((u_int32)(current_time - md->firsttime));
+		im->firsttime = htonl((u_int32)(current_time - md->lasttime));
 		im->lastdrop = htonl((u_int32)md->drop_count);
 		im->count = htonl((u_int32)(md->count));
 		if (md->rmtadr.ss_family == AF_INET6) {
@@ -1901,29 +1961,20 @@ mon_getlist_1(
 {
 	register struct info_monitor_1 *im;
 	register struct mon_data *md;
-	int offset;
 	extern struct mon_data mon_mru_list;
 	extern int mon_enabled;
 
-#ifdef DEBUG
-	if (debug > 2)
-	    printf("wants monitor 1 list\n");
-#endif
 	if (!mon_enabled) {
 		req_ack(srcadr, inter, inpkt, INFO_ERR_NODATA);
 		return;
 	}
-	if (client_v6_capable)
-		offset = 0;
-	else
-		offset = ALT_2ADDR_LONG;
 	im = (struct info_monitor_1 *)prepare_pkt(srcadr, inter, inpkt,
-	    sizeof(struct info_monitor_1) - offset);
+	    v6sizeof(struct info_monitor_1));
 	for (md = mon_mru_list.mru_next; md != &mon_mru_list && im != 0;
 	     md = md->mru_next) {
 		im->lasttime = htonl((u_int32)md->avg_interval);
-		im->firsttime = htonl((u_int32)(current_time - md->firsttime));
-		    im->lastdrop = htonl((u_int32)md->drop_count);
+		im->firsttime = htonl((u_int32)(current_time - md->lasttime));
+		im->lastdrop = htonl((u_int32)md->drop_count);
 		im->count = htonl((u_int32)md->count);
 		if (md->rmtadr.ss_family == AF_INET6) {
 			if (!client_v6_capable)
@@ -1943,7 +1994,7 @@ mon_getlist_1(
 				: GET_INADDR(md->interface->bcast))
 				: 4);
 		}
-		im->flags = md->cast_flags;
+		im->flags = htonl(md->cast_flags);
 		im->port = md->rmtport;
 		im->mode = md->mode;
 		im->version = md->version;
@@ -1991,7 +2042,8 @@ reset_stats(
 	}
 
 	flags = ((struct reset_flags *)inpkt->data)->flags;
-
+	flags = ntohl(flags);
+     
 	if (flags & ~RESET_ALLFLAGS) {
 		msyslog(LOG_ERR, "reset_stats: reset leaves %#lx",
 			flags & ~RESET_ALLFLAGS);
@@ -2237,7 +2289,6 @@ req_get_traps(
 	register struct info_trap *it;
 	register struct ctl_trap *tr;
 	register int i;
-	int offset;
 
 	/*
 	 * Imported from the control module
@@ -2250,12 +2301,8 @@ req_get_traps(
 		return;
 	}
 
-	if (client_v6_capable)
-		offset = 0;
-	else
-		offset = ALT_2ADDR_INT;		
 	it = (struct info_trap *)prepare_pkt(srcadr, inter, inpkt,
-	    sizeof(struct info_trap) - offset);
+	    v6sizeof(struct info_trap));
 
 	for (i = 0, tr = ctl_trap; i < CTL_MAXTRAPS; i++, tr++) {
 		if (tr->tr_flags & TRAP_INUSE) {
@@ -2776,3 +2823,91 @@ get_clkbug_info(
 	flush_pkt();
 }
 #endif
+
+/*
+ * receiver of interface structures
+ */
+static void
+fill_info_if_stats(void *data, interface_info_t *interface_info)
+{
+	struct info_if_stats **ifsp = (struct info_if_stats **)data;
+	struct info_if_stats *ifs = *ifsp;
+	struct interface *interface = interface_info->interface;
+	
+	memset((char*)ifs, 0, sizeof(*ifs));
+	
+	if (interface->sin.ss_family == AF_INET6) {
+		if (!client_v6_capable) {
+			return;
+		}
+		ifs->v6_flag = 1;
+		memcpy((char *)&ifs->unaddr.addr6, (char *)&CAST_V6(interface->sin)->sin6_addr, sizeof(struct in6_addr));
+		memcpy((char *)&ifs->unbcast.addr6, (char *)&CAST_V6(interface->bcast)->sin6_addr, sizeof(struct in6_addr));
+		memcpy((char *)&ifs->unmask.addr6, (char *)&CAST_V6(interface->mask)->sin6_addr, sizeof(struct in6_addr));
+	} else {
+		ifs->v6_flag = 0;
+		memcpy((char *)&ifs->unaddr.addr, (char *)&CAST_V4(interface->sin)->sin_addr, sizeof(struct in_addr));
+		memcpy((char *)&ifs->unbcast.addr, (char *)&CAST_V4(interface->bcast)->sin_addr, sizeof(struct in_addr));
+		memcpy((char *)&ifs->unmask.addr, (char *)&CAST_V4(interface->mask)->sin_addr, sizeof(struct in_addr));
+	}
+	ifs->v6_flag = htonl(ifs->v6_flag);
+	strcpy(ifs->name, interface->name);
+	ifs->family = htons(interface->family);
+	ifs->flags = htonl(interface->flags);
+	ifs->last_ttl = htonl(interface->last_ttl);
+	ifs->num_mcast = htonl(interface->num_mcast);
+	ifs->received = htonl(interface->received);
+	ifs->sent = htonl(interface->sent);
+	ifs->notsent = htonl(interface->notsent);
+	ifs->scopeid = htonl(interface->scopeid);
+	ifs->ifindex = htonl(interface->ifindex);
+	ifs->ifnum = htonl(interface->ifnum);
+	ifs->uptime = htonl(current_time - interface->starttime);
+	ifs->ignore_packets = interface->ignore_packets;
+	ifs->peercnt = htonl(interface->peercnt);
+	ifs->action = interface_info->action;
+	
+	*ifsp = (struct info_if_stats *)more_pkt();
+}
+
+/*
+ * get_if_stats - get interface statistics
+ */
+static void
+get_if_stats(
+	struct sockaddr_storage *srcadr,
+	struct interface *inter,
+	struct req_pkt *inpkt
+	)
+{
+	struct info_if_stats *ifs;
+
+	DPRINTF(3, ("wants interface statistics\n"));
+
+	ifs = (struct info_if_stats *)prepare_pkt(srcadr, inter, inpkt,
+	    v6sizeof(struct info_if_stats));
+
+	interface_enumerate(fill_info_if_stats, &ifs);
+	
+	flush_pkt();
+}
+
+static void
+do_if_reload(
+	struct sockaddr_storage *srcadr,
+	struct interface *inter,
+	struct req_pkt *inpkt
+	)
+{
+	struct info_if_stats *ifs;
+
+	DPRINTF(3, ("wants interface reload\n"));
+
+	ifs = (struct info_if_stats *)prepare_pkt(srcadr, inter, inpkt,
+	    v6sizeof(struct info_if_stats));
+
+	interface_update(fill_info_if_stats, &ifs);
+	
+	flush_pkt();
+}
+
