@@ -363,7 +363,10 @@ TSIP_decode (
 		return 0;	
 	}
 
-	if (up->rpt_buf[0] == (char) 0x8f) {
+	/*
+	 * We cast both to u_char to as 0x8f uses the sign bit on a char
+	 */
+	if ((u_char) up->rpt_buf[0] == (u_char) 0x8f) {
 	/* 
 	 * Superpackets
 	 */
@@ -491,7 +494,7 @@ if (debug > 1) {
 			return 0;
 		}
 
-		pp->nsec = (long) (getdbl((u_char *) &mb(3)) * 1000000);
+		pp->nsec = (long) (getdbl((u_char *) &mb(3)) * 1000000000);
 
 		if ((pp->day = day_of_year(&mb(14))) < 0) 
 			break;
