@@ -3,18 +3,19 @@ repotype=stable
 # odd minor numbers are for -dev, even minor numbers are for -stable
 proto=4
 major=2
-minor=2
+minor=4
 version=${proto}.${major}.${minor}
 # Point.  3 cases:
 # - Numeric values increment
 # - empty 'increments' to 1
 # - NEW 'increments' to empty
-point=4
+point=NEW
 # Special.  Normally unused.  A suffix.
 #special=ag
 special=
 # ReleaseCandidate. 'yes' or 'no'.
 releasecandidate=no
+#releasecandidate=no
 # ChangeLog tag
 CLTAG=NTP_4_2_0
 ###
@@ -23,4 +24,6 @@ CLTAG=NTP_4_2_0
 # - GO triggers a release
 # - - rcpoint gets set to 0
 # - - releasecandidate gets set to no
+# - GRONK is for -dev
 rcpoint=
+#rcpoint=GRONK
